/*
 * (c) 2014 Rechenraum e.U. (Simon Flöry; simon.floery@rechenraum.com)
 *
 * This file is part of project oe3d (www.oe3d.at).
 *
 * oe3d is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * oe3d is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with oe3d. If not, see <http://www.gnu.org/licenses/>.
 */

#include <fstream>
#include <sstream>
#include <iostream>
#include <cmath>

#include "Horizon.h"

//------------------------------------------------------------------------------
 
/*! \brief Constructor.
 * 
 */
Horizon::Horizon()
{

}

//------------------------------------------------------------------------------

/*! \brief Load horizon from file
 *
 * The first line comprises three floating point numbers:
 *      observer_latitude_degree observer_longitude_degree observer_altitude_meters
 * All remaining lines comprise two entries:
 *      azimuth_degee altitude_degree
 * The entries must be ordered from east to west and equally spaced w.r.t.
 * azimuth.
 *
 * \param filename Name of file to load horizon from
 * \param obs Location and altitude will be parsed into this object
 *
 * \return True on success, false otherwise
 */
bool Horizon::load(const std::string& filename, Observer& obs)
{
    std::ifstream ifile;
    ifile.open(filename.c_str());
    if (!ifile.is_open())
    {
        std::cerr << "Horizon::load - Failed to open horizon from " << filename << "." << std::endl;
        return false;
    }

    std::vector< GeCore::GCReal > azimuth;
    std::string line;
    for ( unsigned int lineCount = 0 ; !getline(ifile, line).eof() ; lineCount++ )
    {
        /* this will tokenize the line for us, it uses whitespaces and newlines
         * as separators.
         */
        std::istringstream iss(line);
        GeCore::GCReal value;
        std::vector< GeCore::GCReal > values;
        while (iss >> value)
        {
            values.push_back(value);
        }

        if (lineCount == 0)
        {
            if (values.size() != 3)
            {
                std::cerr << "Horizon::load - Malformatted first line." << std::endl;
                return false;
            }

            obs.m_location.lat = values[0];
            obs.m_location.lng = values[1];
            obs.m_elevation = values[2];
        }
        else
        {
            if (values.size() != 2)
            {
                std::cerr << "Horizon::load - Malformatted line #" << lineCount << "." << std::endl;
                return false;
            }
            azimuth.push_back(values[0]);
            m_altitude.push_back(values[1]);
        }
    }

    ifile.close();

    m_azMin = azimuth.front();
    for ( unsigned int i=1 ; i<azimuth.size() ; i++ )
    {
        while (azimuth[i] < m_azMin)
        {
            azimuth[i] += 360;
        }
    }
    m_azMax = azimuth.back();
    m_azRange = m_azMax - m_azMin;

    return true;
}

//------------------------------------------------------------------------------

/*! \brief Azimuth to bin index conversion.
 * 
 * This method takes an azimuth angle in the configured range of this
 * instance of Horizon and converts it to an index of the altitude array.
 * 
 * \param az Azimuth value in [azMin, azMax]
 * 
 * \return Index to the array, which is in [0,getNrBins()-1]
 */ 
unsigned int Horizon::az2index(const GeCore::GCReal az) const
{
#ifdef SUNNIG_DEBUG
    if (az < m_azMin || az > m_azMax) {
		std::cerr << "Horizon::az2index - az not in [azMin, azMax]" << std::endl;
		return -1;
	}
#endif // SUNNIG_DEBUG
	
    if (az == m_azMin) return 0;
	
    if (az == m_azMax) return m_altitude.size()-1;

    return floor((az - m_azMin) / (m_azMax - m_azMin) * m_altitude.size());
}

//------------------------------------------------------------------------------

/*! \brief Bin index to azimuth conversion.
 *
 * This method takes an array index and converts it to the elements's center
 * azimuth value.
 *
 * \param Index to the array, which is in [0,getNrBins()-1]
 *
 * \return Corresponding element's center azimuth or a negative number if an
 * error occurred. If successful, in [azMin, azMax]
 */
GeCore::GCReal Horizon::index2az(const unsigned int idx) const
{
#ifdef SUNNIG_DEBUG
    if (idx > m_altitude.size()-1) {
		std::cerr << "Horizon::index2az - idx > getNrBins()-1" << std::endl;
        return -1;
    }
#endif
    GeCore::GCReal delta = m_azRange / m_altitude.size();
    return m_azMin + (idx + 0.5)*delta;
}

//------------------------------------------------------------------------------

/*! \brief Get minimal azimuth value
 *
 * See documentation of member variable for more information.
 *
 * \return Minimal azimuth value.
 */
GeCore::GCReal Horizon::getAzMin() const
{
    return m_azMin;
}

//------------------------------------------------------------------------------

/*! \brief Get maximal azimuth value
 *
 * See documentation of member variable for more information.
 *
 * \return Maximal azimuth value.
 */
GeCore::GCReal Horizon::getAzMax() const
{
    return m_azMax;
}

//------------------------------------------------------------------------------

/*! \brief Return const reference to altitude array.
 *
 * \return Const reference to altitude array
 */
const std::vector< GeCore::GCReal >& Horizon::getAltitude() const
{
    return m_altitude;
}

//------------------------------------------------------------------------------
