/*
 * (c) 2014 Rechenraum e.U. (Simon Flöry; simon.floery@rechenraum.com)
 *
 * This file is part of project oe3d (www.oe3d.at).
 *
 * oe3d is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * oe3d is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with oe3d. If not, see <http://www.gnu.org/licenses/>.
 */

#include <iostream>
#include <cmath>

#include <libnova/julian_day.h>
#include <libnova/solar.h>

#include "LabeledSunpath.h"

//------------------------------------------------------------------------------

/*! \brief Constructor
 *
 * The constructor takes a const reference to the options and a pointer to
 * the horizon object and stores them for later use. Apart from this, nothing
 * is done here.
 *
 * \param options Const reference to options object
 * \param horizon Const pointer to computed horizon
 */
LabeledSunpath::LabeledSunpath(const Observer &obs, const Horizon* horizon)
    : Evaluator< bool >(obs), m_horizon(horizon)
{

}

//------------------------------------------------------------------------------

/*! \brief Compute sunshine duration for julian date time span
 *
 * \param jd0 Start julian date, is of type double as libnova returns the julian
 *      date as double. See Options::getLocalJulianDate. Must be local midnight.
 * \param jd1 End julian date, is of type double as libnova returns the julian
 *      date as double. See Options::getLocalJulianDate. Must be local midnight.
 *
 * \return True on success, false otherwise
 */
bool LabeledSunpath::compute(double jd0, double jd1)
{
    m_labeledSunpath = intersect(jd0, jd1);
    return (m_labeledSunpath.size() > 0);
}

//------------------------------------------------------------------------------

/*! \brief Compute labeled sunpath as if there was no mountains.
 *
 * This method is currently not implemented. It should not be too hard to do so.
 *
 * \param jd0 Start julian date, is of type double as libnova returns the julian
 *      date as double. See Options::getLocalJulianDate. Must be local midnight.
 * \param jd1 End julian date, is of type double as libnova returns the julian
 *      date as double. See Options::getLocalJulianDate. Must be local midnight.
 *
 * \return True on success, false otherwise
 */
bool LabeledSunpath::computePlain(double jd0, double jd1)
{
    std::cerr << "LabeledSunpath::computePlain - Currently not implemented." << std::endl;
    return false;
}

//------------------------------------------------------------------------------

/*! \brief Compute the intersected sunpath for the given time range
 *
 * This method combines the sunpath intersection of the single days in the given
 * time range into a single labeled intersected sunpath list.
 *
 * \param jd0 Start julian date, is of type double as libnova returns the julian
 *      date as double. See Options::getLocalJulianDate. Must be local midnight.
 * \param jd1 End julian date, is of type double as libnova returns the julian
 *      date as double. See Options::getLocalJulianDate. Must be local midnight.
 *
 * \return Labeled and intersected sunpath
 */
std::list< LabeledSunpath::spSampleT > LabeledSunpath::intersect(double jd0, double jd1) const
{
    /* one object for all */
    Sunpath sp(m_obs, m_horizon->getAzMin(), m_horizon->getAzMax());

    /* cumulated intersection info */
    std::list< spSampleT > s;

    /* as we require jd1 to be local midnight, it is safe to subtract some
     * offset to have a proper stopping condition (We compute the sunpath
     * day-wise).
     */
    for ( double jd = jd0 ; jd < jd1-0.1 ; jd += 1. ) {
        /* compute sun path for this day. Again, as jd is midnight, add some eps
         * to clearly be in the wanted day.
         */
        if (!sp.compute(jd + 0.1)) return s;

        /* intersect this sunpath with the horizon */
        std::list< spSampleT > sLocal = intersect(*m_horizon, sp);

        /* If we have a temporal overlap between s and sLocal, we remove the
         * tail of s. This is the case if the sun does not set, or does nearly
         * not set. Recall that sp.compute returns an extra sample in the
         * beginning and end to have a sure intersect with the horizon.
         */
        if (s.size() > 0)
            while (s.back().jd > sLocal.front().jd) s.pop_back();

        /* combine sLocal and s */
        s.insert(s.end(), sLocal.begin(), sLocal.end());
    }

    return s;
}

//------------------------------------------------------------------------------

/*! \brief Compute effective sunshine duration for given horizon and sun path
 *
 * This computes line-line intersections. You will most likely call compute on
 * the result.
 *
 * \param h Horizon
 * \param sp Sunpath
 *
 * \return Sunpath labelled with intersection info.
 */
std::list< LabeledSunpath::spSampleT > LabeledSunpath::intersect(const Horizon& h, const Sunpath& sp) const
{
    std::list< spSampleT > s;

    /* Convert sunpath to a labelled representation */
    const GeCore::GCPolyline& s_pl = sp.get();
    GeCore::GCPolyline::const_iterator s_plIt = s_pl.begin();
    for ( ; s_plIt != s_pl.end() ; ++s_plIt ) {
        spSampleT spSample;
        const GeCore::GC3D& p3d = *s_plIt;
        spSample.p = GeCore::GC2D(p3d.x, p3d.y);
        spSample.jd = p3d.z;
        spSample.label = UNKNOWN;
        s.push_back(spSample);
    }

    const std::vector< GeCore::GCReal >& h_altitude  = h.getAltitude();

    /*
     * compute all intersections on sun path
     */

    /* GCREAL_EPS is 1e-8 which is rather small for the intersection tests. so we
     * define our own eps.
     */
    GeCore::GCReal eps = 1e-8;

    std::list< spSampleT >::iterator sIt = s.begin();
    for ( sIt++ ; sIt != s.end() ; sIt++ ) {
        /* get iterator to previous element */
        std::list< spSampleT >::iterator sItPrev = sIt;
        std::advance(sItPrev, -1);

        /* the sunpath's 2d line segment is given by a0, a1 */
        const GeCore::GC2D& a0 = sItPrev->p;
        const GeCore::GC2D& a1 = sIt->p;

        /* prepare for bbox tests later on */
        GeCore::GCReal minAy = a0.y;
        GeCore::GCReal maxAy = a1.y;
        if (minAy > maxAy) std::swap(minAy, maxAy);

        /* So the current segment of the path is a0,a1. Find corresponding
         * part on the horizon by finding the bins for pl0.az and pl1.az.
         * Ensure that i+1 exists. Use i0-1 to have a proper overlap.
         */
        unsigned int i0 = h.az2index(a0.x);
        if (i0 > 0) i0--;
        unsigned int i1 = h.az2index(a1.x);
        if (i1 >= h_altitude.size()-1) i1 = h_altitude.size()-2;

        /* process horizon segment from left to right */
        GeCore::GC2D b1 = GeCore::GC2D(h.index2az(i0), h_altitude[i0]);
        for ( unsigned int i=i0 ; i<i1 ; i++ ) {
            GeCore::GC2D b0 = b1;
            b1 = GeCore::GC2D(h.index2az(i+1), h_altitude[i+1]);

            /* intersect a0a1 and b0b1. first a bbox test. */
            if (std::min(b0.y, b1.y) > maxAy+eps) continue;
            if (std::max(b0.y, b1.y) < minAy-eps) continue;

            /* then the intersection tests by orientation */
			GeCore::GCReal aab0 = GeCore::computeSignedArea(a0, a1, b0);
			GeCore::GCReal aab1 = GeCore::computeSignedArea(a0, a1, b1);
			GeCore::GCReal bba0 = GeCore::computeSignedArea(b0, b1, a0);
			GeCore::GCReal bba1 = GeCore::computeSignedArea(b0, b1, a1);

            /* handle all possible cases */
            if (aab0 * aab1 > eps || bba0 * bba1 > eps) {
                /* one segment is located completely to one side of the other
                 * segment. So there is no intersection.
                 */
            } else if (aab0 * aab1 < (-1)*eps && bba0 * bba1 < (-1)*eps) {
                /* this is the proper intersection case */
                GeCore::GC2D v = a1 - a0;
                GeCore::GC2D w = b0 - b1;
                GeCore::GC2D rhs = b0 - a0;
                GeCore::GCReal lambda = (rhs.x*w.y-rhs.y*w.x) / (v.x*w.y - v.y*w.x);
#ifdef SUNNIG_DEBUG
                if (lambda < eps || lambda > 1-eps)
                    std::cerr << "LabeledSunpath::compute - Numerical inprecisions in proper intersection case." << std::endl;
#endif // SUNNIG_DEBUG
                /* insert a0 + lambda v between a0 and a1 */
                sIt = insert(s, sItPrev, sIt, lambda);
                break;
            } else if (aab0 * aab1 < (-1)*eps) {
                /* so there is an intersection with one vertex of a0a1 on b0b1.
                 * As in this case there is no need to insert a vertex to a0a1,
                 * we only label the vertices as being on the horizon.
                 */
                if (fabs(bba0) < eps) {
                    /* vertex a0 is on b0b1. This case was done in the previous
                     * iteration. Do nothing now.
                     */
                } else {
                    /* vertex a1 is on b0b1. update label. */
                        sIt->label = ON_HORIZON;
                }
            } else if (bba0 * bba1 < (-1)*eps) {
                /* so there is an intersection with one vertex of b0b1 on a0a1 */
                if (fabs(aab0) < eps) {
                    /* vertex b0 is on a0a1. This case was done in the previous
                     * iteration. Do nothing now.
                     */
                } else {
                    /* vertex b1 is on a0a1. insert b1 between a0 and a1. */
                    GeCore::GCReal lambda = (b1.x - a0.x) / (a1.x - a0.x);
#ifdef SUNNIG_DEBUG
                    if (lambda < eps || lambda > 1-eps)
                        std::cerr << "LabeledSunpath::compute - Numerical inprecisions in b1 on a0a1." << std::endl;
#endif // SUNNIG_DEBUG
                    sIt = insert(s, sItPrev, sIt, lambda);
                    break;
                }
            } else {
                /* the two segments are colinear. only do something if b1 is
                 * between a0a1.
                 */
                GeCore::GCReal lambda = (b1.x - a0.x) / (a1.x - a0.x);
                if (lambda > eps && lambda < 1-eps) {
                    sIt = insert(s, sItPrev, sIt, lambda);
                    break;
                }
            }
        }
    }

    /*
     * label all sunpath vertices which are still in unknown position to the
     * horizon. We "simply" need to determine if the horizon is passing above
     * or below the sunpath vertex.
     */

    sIt = s.begin();
    for ( ; sIt != s.end() ; sIt++) {
        if (sIt->label == UNKNOWN) {
            const GeCore::GCReal& sx = sIt->p.x;
            const GeCore::GCReal& sy = sIt->p.y;
            /* get bin for sunpath vertex (sx,sy) */
            int idx0 = h.az2index(sx);
            GeCore::GCReal az0 = h.index2az(idx0);
            GeCore::GCReal altRef = GCREAL_MAX;
            if (fabs(az0 - sx) < eps) {
                /* If sunpath vertex is close to the bin center, take the bin's
                 *  altitutde
                 */
                altRef = h_altitude[idx0];
            } else {
                /* else we need to take the horizon segment and interpolate its
                 * altitude at the azimuth of the sunpath vertex.
                 */
                if (az0 > sx) {
                    /* Sunpah vertex is left of the bin's center */
                    if (idx0 <= 0) {
                        /* no chance to interpolate in this case */
                        altRef = h_altitude.front();
                    } else {
                        /* idx0 will be the start index of the horizon segment
                         * which is the one to the left.
                         */
                        idx0--;
                    }
                } else {
                    /* Sunpath vertex is right of the bin's center */
                    if (idx0 >= (int) h_altitude.size()-1) {
                        /* If we are at the end there is no chance to interpolate. */
                        altRef = h_altitude.back();
                    }
                }
                /* If we still want to interpolate (e.g. there was no special case) */
                if (altRef == GCREAL_MAX) {
                    az0 = h.index2az(idx0);
                    GeCore::GCReal az1 = h.index2az(idx0+1);
                    GeCore::GCReal lambda = (sx - az0) / (az1 - az0);
#ifdef SUNNIG_DEBUG
                    if (lambda < eps || lambda > 1-eps)
                        std::cerr << "LabeledSunpath::compute - Numerical inprecisions in labeling." << std::endl;
#endif // SUNNIG_DEBUG
                    altRef = h_altitude[idx0] + lambda*(h_altitude[idx0+1] - h_altitude[idx0]);
                }
            }

            /* now label depending on the reference altitude */
            if (fabs(altRef - sy) < eps) sIt->label = ON_HORIZON;
            else if (sy > altRef) sIt->label = ABOVE_HORIZON;
            else sIt->label = BELOW_HORIZON;
        }
    }

    return s;
}

//------------------------------------------------------------------------------

/*! \brief Insert linearly interpolated sample into sunpath sample list
 *
 * \param s Sunpath samples
 * \param sItPrev Iterator pointing to first element
 * \param sIt Iterator pointing to sItPrev+1
 * \param lambda Value in [0,1] to interpolate between sItPrev and sIt
 *
 * \return Iterator pointing to the newly inserted element.
 */
std::list< LabeledSunpath::spSampleT >::iterator LabeledSunpath::insert(
        std::list< spSampleT >& s,
        std::list< spSampleT >::iterator sItPrev,
        std::list< spSampleT >::iterator sIt,
        GeCore::GCReal lambda) const
{
    spSampleT spSample;
    spSample.p  = sItPrev->p  + (sIt->p  - sItPrev->p ) * lambda;
    spSample.jd = sItPrev->jd + (sIt->jd - sItPrev->jd) * lambda;
    spSample.label = ON_HORIZON;
    return s.insert(sIt, spSample);
}

//------------------------------------------------------------------------------

/*! \brief Obtain const reference to list of sunpath labels
 *
 * The list is sorted by time and may not be assumed connected. There will be
 * gaps at night.
 *
 * First coordinate is azimuth (in DEGREES), second altitude (in DEGREES).
 *
 * \return Const reference to list of sunpath labels.
 */
const std::list< LabeledSunpath::spSampleT >& LabeledSunpath::getSunpath() const
{
    return m_labeledSunpath;
}

//------------------------------------------------------------------------------

/*! \brief Get Julian start date the labeled sunpath for computed for.
 *
 * \return Julian start date of labeled sunpath or a date in the future if there
 *      is no sunpath
 */
double LabeledSunpath::getJulianDateStart() const
{
    return m_labeledSunpath.size() == 0 ? GCREAL_MAX : m_labeledSunpath.front().jd;
}

//------------------------------------------------------------------------------

/*! \brief Get Julian start end the labeled sunpath for computed for.
 *
 * \return Julian end date of labeled sunpath or a date in the future if there
 *      is no sunpath
 */
double LabeledSunpath::getJulianDateEnd() const
{
    return m_labeledSunpath.size() == 0 ? GCREAL_MAX : m_labeledSunpath.back().jd;
}

//------------------------------------------------------------------------------

/*! \brief Check if the given timestamp is covered by this sunpath object
 *
 * \begin jd0 Begin of time span as Julian date
 * \begin jd1 End of time span as Julian date
 *
 * \return True, if time span is covered and false otherwise.
 */
bool LabeledSunpath::covers(double jd0, double jd1) const
{
    if (jd0 < getJulianDateStart()) return false;
    if (jd1 > getJulianDateEnd()) return false;
    return true;
}

//------------------------------------------------------------------------------
