/*
 * (c) 2014 Rechenraum e.U. (Simon Flöry; simon.floery@rechenraum.com)
 *
 * This file is part of project oe3d (www.oe3d.at).
 *
 * oe3d is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * oe3d is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with oe3d. If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef LABELED_SUNPATH_H
#define LABELED_SUNPATH_H

#include <list>

#include <libnova/ln_types.h>

/* forward declaration */
class LabeledSunpath;

#include "sunnig.h"
#include "Observer.h"
#include "Horizon.h"
#include "Sunpath.h"
#include "evaluator/Evaluator.h"

/*! \brief Result of intersection of sunpath with horizon
 *
 */
class LabeledSunpath : public Evaluator< bool >
{
public:
    /* calls to members declared only in a base template class are not resolved
     * when they are encountered, as their might be specializations that are
     * found later. So we end up with compute and computePlain unknown, unless
     * we explicitely bring them into scope with the following using
     * declarations. See
     *      http://womble.decadent.org.uk/c++/template-faq.html#base-lookup
     * for more information.
     */
    using Evaluator< bool >::compute;
    using Evaluator< bool >::computePlain;

    /*! This labels the samples of the intersected sunpath */
    enum plLabelT {
        UNKNOWN,            /*!< Sample is not labeled */
        BELOW_HORIZON,      /*!< Sample is not visible, e.g. below horizon */
        ON_HORIZON,         /*!< Sample is on the horizon, e.g. an intersection point of the sunpath and the horizon */
        ABOVE_HORIZON       /*!< Sample is visible, e.g. above horizon */
    };

    /*! This is a sample of the labeled sunpath */
    struct spSampleT {
        GeCore::GC2D p;         /*!< Coordiantes of sample, first coordinate is azimuth (in DEGREES), second altitude (in DEGREES)
                                 *      azimuth is 0=S, 90=W, 180=N, 270=E
                                 *      altitude is 0 on the horizon and 90 degrees for the zenit
                                 */
        GeCore::GCReal jd;      /*!< Julian date of sample (estimate) */
        plLabelT label;         /*!< Label indicating position of sample w.r.t. horizon */
    };

    /* documented in source code */
    LabeledSunpath(const Observer&, const Horizon*);

    /* documented in source code */
    bool compute(double, double);

    /* documented in source code */
    bool computePlain(double, double);

    /* documented in source code */
    const std::list< spSampleT >& getSunpath() const;

    /* documented in source code */
    double getJulianDateStart() const;
    double getJulianDateEnd() const;

    /* documented in source code */
    bool covers(double, double) const;

private:
    /* documented in source code */
    std::list< spSampleT > intersect(double, double) const;

    /* documented in source code */
    std::list< spSampleT > intersect(const Horizon&, const Sunpath&) const;

    /* documented in source code */
    std::list< spSampleT >::iterator insert(
            std::list< spSampleT >&,
            std::list< spSampleT >::iterator,
            std::list< spSampleT >::iterator,
            GeCore::GCReal) const;

    /*! private pointer to the computed horizon */
    const Horizon* m_horizon;

    /*! The labeled sunpath */
    std::list< spSampleT > m_labeledSunpath;
};

#endif // LABELED_SUNPATH_H
