/*
 * (c) 2014 Rechenraum e.U. (Simon Flöry; simon.floery@rechenraum.com)
 *
 * This file is part of project oe3d (www.oe3d.at).
 *
 * oe3d is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * oe3d is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with oe3d. If not, see <http://www.gnu.org/licenses/>.
 */

#include <iostream>
#include <cstdlib>

#include <libnova/utility.h>
#include <libnova/julian_day.h>
#include <libnova/solar.h>
#include <libnova/transform.h>

#include "Observer.h"

//------------------------------------------------------------------------------
 
/*! \brief Constructor.
 * 
 */
Observer::Observer()
{

}

//------------------------------------------------------------------------------

/*! \brief Get observer's offset from GMT in time seconds
 *
 * For, ln_get_julian_local_date we need to convert from arc seconds w.r.t.
 * longitude to seconds w.r.t. to time. Our longitude is in [-180, 180],
 * whereas the gmtoff argument defining the number of time seconds east
 * of GMT is in [0...24*60*60=86400].
 *
 * \return Offset in time seconds, 0 ... 86400
 */
double Observer::getGMTOffset() const
{
    GeCore::GCReal gmtoff = m_location.lng;
    if (gmtoff < 0)
    {
        gmtoff += 360;
    }
    return gmtoff / 360. * 24. * 3600.;
}

//------------------------------------------------------------------------------

/*! \brief Convert given local date to local Julian date
 *
 * A Julian day starts at noon Greenwhich Mean Time. Hence, we need first to be
 * aware of that the integer part of a Julian day tells nothing about the date
 * and second, when given a local time, we need to convert the time to GMT.
 * This is done in this method based on the observers coordinates.
 *
 * \param years Local year (may be negative)
 * \param months Local month
 * \param days Local day
 * \param hours Local hour
 * \param minutes Local minute
 * \param seconds Local second
 *
 * \return Julian Day in GMT
 */
double Observer::getLocalJulianDate(
        int years,
        int months,
        int days,
        int hours,
        int minutes,
        double seconds) const
{
#ifdef SUNNIG_DEBUG
    if (months < 1 || months > 12) {
        std::cerr << "Options::getLocalJulianDate - Month " << months << " out of expected [1,12] range." << std::endl;
        return -1;
    }

    if (days < 1 || days > 31) {
        std::cerr << "Options::getLocalJulianDate - Days " << days << " out of expected [1,31] range." << std::endl;
        return -1;
    }

    /* We allow 24 to specify 24:00 as end of the day */
    if (hours < 0 || hours > 24) {
        std::cerr << "Options::getLocalJulianDate - Hours " << hours << " out of expected [0,24] range." << std::endl;
        return -1;
    }

    /* We allow 60 to specify :60 as end of the hour */
    if (minutes < 0 || minutes > 60) {
        std::cerr << "Options::getLocalJulianDate - Minutes " << minutes << " out of expected [0,60] range." << std::endl;
        return -1;
    }

    /* We allow 60 to specify :60 as end of the minute */
    if (seconds < 0 || seconds > 60) {
        std::cerr << "Options::getLocalJulianDate - Seconds " << seconds << " out of expected [0,60] range." << std::endl;
        return -1;
    }
#endif // SUNNIG_DEBUG

    ln_zonedate queryTime = { years, months, days, hours, minutes, seconds, (long) getGMTOffset()};

    return ln_get_julian_local_date(&queryTime);
}

//------------------------------------------------------------------------------


/*! \brief Convert Julian date to local time based on observer position
 *
 * \param jd Julian day
 *
 * \return Local date
 */
ln_date Observer::getLocalDate(double jd) const
{
    ln_date gmtDate;
    ln_get_date(jd, &gmtDate);

    ln_zonedate localZoneDate;
    ln_date_to_zonedate(&gmtDate, &localZoneDate, (long) getGMTOffset());

    ln_date localDate = {
        localZoneDate.years,
        localZoneDate.months,
        localZoneDate.days,
        localZoneDate.hours,
        localZoneDate.minutes,
        localZoneDate.seconds
    };

    return localDate;
}

//------------------------------------------------------------------------------


/*! \brief Get solar rise/set/transit time on given Julian day
 *
 * For us, a Julian day represents a calendar date and we want the rise and set
 * time on that calendar day. For libnova, a Julian day is - well - a Julian day
 * and it returns the rise time on the given Julian day. This gives a big mess
 * all around, so we wrote this method, which garantuees, that rise and set are
 * on the same calender day as the given Julian day jd.
 *
 * \param jd Julian day, that we interprete as local calendar date
 * \param rst Rise/set/transit structure; rise and set are on the same local
 *      calender date as the argument jd (if there is any rise/set)
 *
 * \return Just as ln_get_solar_rst (and ln_get_body_rst_horizon), this method
 *      returns 0 if there is a proper rise/set time, -1 if the sun is below
 *      the horizon the entire day and 1, if it is above the horizon all day.
 */
int Observer::getSolarRST(double jd, ln_rst_time& rst) const
{
    /* create non-const copy */
    ln_lnlat_posn obsPosP = m_location;

    ln_date date = getLocalDate(jd);
    double jd0 = getLocalJulianDate(date.years, date.months, date.days, 0, 0, 0);
    double jd1 = getLocalJulianDate(date.years, date.months, date.days, 24, 0, 0);

    int status = ln_get_solar_rst(jd, &obsPosP, &rst);
    if (status == 0) {
        /* fix rise time if necessary */
        ln_rst_time rst_rtmp;
        int status_rtmp = 2;
        /* if rise time is smaller than requested day's beginning, get rise time
         * for next day. Same argument if rise time is past requested day.
         */
        if (rst.rise < jd0) status_rtmp = ln_get_solar_rst(jd+1, &obsPosP, &rst_rtmp);
        else if (rst.rise > jd1) status_rtmp = ln_get_solar_rst(jd-1, &obsPosP, &rst_rtmp);
        /* If there is a new valid rise time, write it to rst */
        if (status_rtmp == 0) rst.rise = rst_rtmp.rise;
        /* If the update computation turned out a special case, we return this
         * special case. We know that the update computation will yield the
         * correct rise time, and if there is a special case, there is no rise
         * time.
         */
        if (abs(status_rtmp) == 1) return status_rtmp;

        /* same for set time */
        ln_rst_time rst_stmp;
        int status_stmp = 2;
        if (rst.set < jd0) status_stmp = ln_get_solar_rst(jd+1, &obsPosP, &rst_stmp);
        else if (rst.rise > jd1) status_stmp = ln_get_solar_rst(jd-1, &obsPosP, &rst_stmp);
        if (status_stmp == 0) rst.set = rst_stmp.set;
        if (abs(status_stmp) == 1) return status_rtmp;

        /* return status 0 */
        return 0;
    } else {
        /* We have either status=1 or status=-1. Try previous and next day for
         * valid rise/set times and return these if found. Otherwise, return
         * special status.
         */
        int statusPre = ln_get_solar_rst(jd-1, &obsPosP, &rst);
        if (statusPre != status) {
            /* If != 0, the status can't have changed */
            if (statusPre != 0) {
                std::cerr << "Options::getSolarRST - We flipped in prev from 1 to -1 or vice versa." << std::endl;
                return status;
            }
            /* If we have fine rise/set times, return these. */
            if (rst.rise >= jd0 && rst.rise <= jd1 && rst.set >= jd0 && rst.set <= jd1) return 0;
        }

        int statusPost = ln_get_solar_rst(jd+1, &obsPosP, &rst);
        if (statusPost != status) {
            if (statusPost != 0) {
                std::cerr << "Options::getSolarRST - We flipped in prev from 1 to -1 or vice versa." << std::endl;
                return status;
            }
            if (rst.rise >= jd0 && rst.rise <= jd1 && rst.set >= jd0 && rst.set <= jd1) return 0;
        }
    }


    return 4;
}

//------------------------------------------------------------------------------
