/*
 * (c) 2014 Rechenraum e.U. (Simon Flöry; simon.floery@rechenraum.com)
 *
 * This file is part of project oe3d (www.oe3d.at).
 *
 * oe3d is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * oe3d is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with oe3d. If not, see <http://www.gnu.org/licenses/>.
 */

#include <iostream>

#include <math.h>

#include <libnova/julian_day.h>
#include <libnova/solar.h>
#include <libnova/transform.h>
#include <libnova/sidereal_time.h>

#include "Sunpath.h"

//------------------------------------------------------------------------------

/*! \brief Constructor
 * 
 * \param opt Constant reference to main options object
 */
Sunpath::Sunpath(const Observer& obs, const GeCore::GCReal azMin, const GeCore::GCReal azMax) :
    m_obs(obs),
    m_azMin(azMin),
    m_azMax(azMax),
    m_sunpathResolution(30)
{

}

//------------------------------------------------------------------------------

/*! \brief Compute the sun path for a given day
 * 
 * This method is the class' working horse. It computes the sun positions in a
 * horizontal coordinate system with the observer's location as origin at the
 * configured number of samples.
 *
 * \param jd Julian date of day to compute sunpath for.
 *
 * \return Samples of the computed sunpath, first coordinate is azimuth, second
 *		cos(altitude), third Julian date of sample. The second and the second to
 *      last sample shall be sun rise and sun set, resp. The extra sample in
 *      the beginnind and end is there to have a sure intersect with the horizon.
 */
bool Sunpath::compute(double jd)
{
    return _computeLibnovaExact(jd);
}

//------------------------------------------------------------------------------

/*! \brief Compute the sun path for a given day based on libnova
 *
 * This method computes the sunpath for a given day with the help of libnova.
 * This approach is considered pretty accurate. We compute sun rise and set,
 * which defines the time interval we are interested in. Then, we compute the
 * sun position with ln_get_solar_equ_coords in the configured number of samples
 * which is converted to horizontal coordinates finally.
 *
 * Be aware that a Julian day starts at noon!
 *
 * \param jd Julian date of day to compute sunpath for. The Julian day
 *      corresponds to a local date, and on this the sunpath is computed.
 *
 * \return True on success, false otherwise.
 */
bool Sunpath::_computeLibnovaExact(double jd)
{
    /* we compute sun rise and sun set. */
	ln_rst_time rst;
    int specialCases = m_obs.getSolarRST(jd, rst);
    double sunRise = rst.rise;
    double sunSet = rst.set;
	if (specialCases == 1) {
        /* the sun is above the horizon the entire day, set rise and set to
         * midnight, resp.
         */
        ln_date date = m_obs.getLocalDate(jd);
        sunRise = m_obs.getLocalJulianDate(date.years, date.months, date.days, 0, 0, 0);
        sunSet = sunRise + 1;
    } else if (specialCases == -1) {
        /* the sun is below the horizon the entire day */
        m_computedPath.clear();
        return true;
    }

#if 0
    GCOut() << "Sunpath::_computeLibnovaExact: " << std::fixed << jd << " = " << m_options.getLocalDate(jd) << std::endl;
    GCOut() << "Sunpath::_computeLibnovaExact: " << std::fixed << sunRise << " = " << m_options.getLocalDate(sunRise) << std::endl;
    GCOut() << "Sunpath::_computeLibnovaExact: " << std::fixed << sunSet << " = " << m_options.getLocalDate(sunSet) << std::endl;
#endif

#ifdef SUNNIG_DEBUG
    if (sunSet < sunRise) {
		std::cerr << "Sunpath::_computeLibnovaExact - Sun rise is after sun set." << std::endl;
        return false;
    }
#endif // SUNNIG_DEBUG

    /* now as we have the bounds, compute the number of samples we get. We want
     * one sample before sun rise and one after sun set. The second and the
     * second to last sample shall be sun rise and sun set, resp.
     */

    unsigned int nrSamples = ceil((sunSet - sunRise) / (m_sunpathResolution / (24. * 60.))) + 2;
//    unsigned int nrSamples = ceil((sunSet - sunRise) / (m_options.sunpathResolution / (24. * 60.)));
    m_computedPath.clear();
    m_computedPath.reserve(nrSamples);
    double jdStep = (sunSet - sunRise) / (nrSamples - 3);
//    double jdStep = (sunSet - sunRise) / (nrSamples - 1);
    jd = sunRise - jdStep;
//    jd = sunRise;
    /* non const copy of observer position */
    ln_lnlat_posn obsPos = m_obs.m_location;
    for (unsigned int i=0; i<nrSamples; i++, jd += jdStep) {
        GeCore::GC3D p;
        if (!_computeLibnovaExact(&obsPos, jd, p)) return false;
		m_computedPath.push_back(p);
    }

#if 0
    if (m_computedPath.front()[0] < m_options.azMin || m_computedPath.back()[0] > m_options.azMax) {
		std::cerr << "Sunpath::_computeLibnovaExact - Sunpath exceeds az bounds in options object: "
                    << m_computedPath.front()[0] << " < " << m_options.azMin << " or "
                    << m_computedPath.back()[0] << " > " << m_options.azMax << "." << std::endl;
		return false;
	}
#endif

	return true;
}

//------------------------------------------------------------------------------

/*! \brief Compute sun position for given time and position on earth
 *
 * This low-level method computes the sun's position in horizontal coordinates
 * for the observer's position: First coordinate is azimuth, second sin(altitude)
 * and third Julian date of sample.
 *
 * \param obsPos Observer's position
 * \param jd Julian date to compute sun position for
 * \param p (out) Sun#s position in horizontal coordinates, centered at the
 *      observer.
 *
 * \return True on success, false otherwise.
 */
bool Sunpath::_computeLibnovaExact(ln_lnlat_posn* obsPos, double jd, GeCore::GC3D& p) const
{
    /* get the position of the sun at Julian date jd in apparent equatorial
     * coordinates.
     */
    ln_equ_posn equ;
    ln_get_solar_equ_coords(jd, &equ);

    /* Convert the apparent equatorial to horizontal coordinates. */
    ln_hrz_posn hrzCoords;
    ln_get_hrz_from_equ(&equ, obsPos, jd, &hrzCoords);

    /* and store the horizontal coordinates along with the Julian date. We
     * have to convert the altitude from degrees to radians before computing
     * the sine. alt is 0 for the horizon and 90 degrees for the zenit, hence
     * it is the sine we need to take.
     * The azimuth may need addition by 360 in certain cases.
     */
    if (m_azMax > 360 && hrzCoords.az < 0.5*(m_azMin + m_azMax) - 180)
    {
        hrzCoords.az += 360;
    }
    p = GeCore::GC3D(hrzCoords.az, hrzCoords.alt, jd);

    return true;
}

//------------------------------------------------------------------------------

/*! \brief Compute today's sunpath
 *
 * This is a convenience method to compute today's (according to system time)
 * sunpath. We obtain today's Julian date with ln_get_julian_from_sys and
 * call compute().
 *
 * \return Samples of the computed sunpath, first coordinate is azimuth, second
 *		cos(altitude), third Julian date of sample. It is feasible to compute
 *		the day length from back()[2] - front()[2].
 */
bool Sunpath::computeToday()
{
    return compute(ln_get_julian_from_sys());
}

//------------------------------------------------------------------------------

/*! \brief Compute the sun path at the summer solstice
 *
 * \param year Year to compute sunpath for
 *
 * \return True on success, false otherwise.
 */
bool Sunpath::computeSummerSolstice(int year)
{
    return compute(m_obs.getLocalJulianDate(year, 6, 21, 12, 0, 0));
}

//------------------------------------------------------------------------------

/*! \brief Compute the sun path at the winter solstice
 *
 * \param year Year to compute sunpath for
 *
 * \return True on success, false otherwise.
 */
bool Sunpath::computeWinterSolstice(int year)
{
    return compute(m_obs.getLocalJulianDate(year, 12, 21, 12, 0, 0));
}

//------------------------------------------------------------------------------

/*! \brief Compute the analemma, that is the sun's position for the same day
 *      time over a year.
 *
 * An analemma is that curve that we obtain when we compute the sun's position
 * at the same time of the day over the course of a year, see also
 *      http://de.wikipedia.org/wiki/Analemma
 *
 * We compute it as closed polyline by repeating the value for 1.1. in the end.
 *
 * \param jd Julian date. This argument specifies both the year of computation
 *      and the time of the day. Day and month are ignored.
 *
 * \return True on success, false otherwise.
 */
bool Sunpath::computeAnalemma(double jd)
{
    ln_date date = m_obs.getLocalDate(jd);
    double time0 = m_obs.getLocalJulianDate(
                date.years, 1, 1,
                date.hours, date.minutes, date.seconds);

    unsigned int nrSamples = m_obs.getLocalJulianDate(
                date.years+1, 1, 1,
                date.hours, date.minutes, date.seconds) - time0;

#ifdef SUNNIG_DEBUG
    if ((nrSamples != 365) && (nrSamples != 366)) {
		std::cerr << "Sunpath::computeAnalemma - Unexpected number of " << nrSamples << " days for the year " << date.years << "." << std::endl;
        return false;
    }
#endif // SUNNIG_DEBUG

    m_computedPath.clear();
    m_computedPath.reserve(nrSamples + 1);

    /* create non-const copy */
    ln_lnlat_posn obsPos = m_obs.m_location;
    for ( unsigned int i = 0 ; i<nrSamples ; i++ ) {
        GeCore::GC3D p;
        if (!_computeLibnovaExact(&obsPos, time0 + i, p)) return false;
		m_computedPath.push_back(p);
    }
    /* make it a closed curve */
	m_computedPath.push_back(m_computedPath.front());

    return true;
}

//------------------------------------------------------------------------------

/*! \brief Return const reference to computed sunpath polyline
 *
 * First coordinate is azimuth (in DEGREES), second altitude (in DEGREES),
 * third (an estimate of) the Julian date of the sample.
 *
 * \return Const reference to computed sunpath polyline.
 */
const GeCore::GCPolyline& Sunpath::get() const
{
    return m_computedPath;
}

//------------------------------------------------------------------------------

/*! \brief Return sun path as ordered list of points
 *
 * \return List
 */
std::list< GeCore::GC3D > Sunpath::getAsList() const
{
    std::list< GeCore::GC3D > l;

    for ( unsigned int i=0 ; i<m_computedPath.size() ; i++ )
	{
		l.push_back(m_computedPath[i]);
	}

    return l;
}

//------------------------------------------------------------------------------
