/*
 * (c) 2014 Rechenraum e.U. (Simon Flöry; simon.floery@rechenraum.com)
 *
 * This file is part of project oe3d (www.oe3d.at).
 *
 * oe3d is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * oe3d is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with oe3d. If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef _SUNPATH_H
#define _SUNPATH_H

#include <list>

#include "geometry/GCPolyline.h"

#include "sunnig.h"
#include "Observer.h"

/*! \brief Compute sun path for given time frame in horizontal coordinate system
 * 
 * This method computes the path of the sun over a given day in a user-specified
 * resolution. The sun path is represented in
 * a horizontal coordiante system with the observer's location as origin. At 
 * least in its current implementation, this class relies on libnova 
 * (http://lobnova.sf.net) for the astronomical part of the computations.
 *
 * It makes sense to compute the sunpath over an entire day only. The sun either
 * rises and sets over the course of a day, in which case it makes sense to
 * compute the sunshine duration daywise. If the sun does not set, the sun
 * wraps in our horizontal coordinate system. These wraps in the horizontal
 * coordinate system shall be aligned to local midnight, and thus computing the
 * sunpath day-wise makes sense again.
 */
class Sunpath
{
public:
	/* documented in source file */
    Sunpath(const Observer&, const GeCore::GCReal, const GeCore::GCReal);

	/* documented in source file */
	bool compute(double);

	/* documented in source file */
	bool computeToday();

	/* documented in source file */
	bool computeSummerSolstice(int);

	/* documented in source file */
	bool computeWinterSolstice(int);

    /* documented in source file */
    bool computeAnalemma(double);

    /* documented in source file */
    const GeCore::GCPolyline& get() const;

    /* documented in source file */
    std::list< GeCore::GC3D > getAsList() const;

	
private:

    /**************************************************************************
     * Private methods
     **************************************************************************/

    /* documented in source file */
	bool _computeLibnovaExact(double);

    /* documented in source file */
    bool _computeLibnovaExact(ln_lnlat_posn*, double, GeCore::GC3D&) const;

    /**************************************************************************
     * Private members
     **************************************************************************/

    /*! Our private copy of the observer object */
    const Observer m_obs;

    /*! azMin of Horizon, we need it in _computeLibnovaExact */
    const GeCore::GCReal m_azMin;

    /*! azMax of Horizon, we need it in _computeLibnovaExact */
    const GeCore::GCReal m_azMax;

    /*! Samples of the computed sunpath, first coordinate is azimuth (in
     * DEGREES), second altitude (in DEGREES), third (an estimate of) the
     * Julian date of the sample.
	 */
    GeCore::GCPolyline m_computedPath;

    /*! Sunpath resolution in minutes */
    const GeCore::GCReal m_sunpathResolution;
};

#endif // _SUNPATH_H
