/*
 * (c) 2014 Rechenraum e.U. (Simon Flöry; simon.floery@rechenraum.com)
 *
 * This file is part of project oe3d (www.oe3d.at).
 *
 * oe3d is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * oe3d is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with oe3d. If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef EVALUATOR_H
#define EVALUATOR_H

/* forward declaration */
template< typename T > class Evaluator;

#include "../sunnig.h"
#include "../Observer.h"

//------------------------------------------------------------------------------

/******************************************************************************
 *
 * class Evaluator< T >
 *
 ******************************************************************************/

//------------------------------------------------------------------------------

/*! \brief Abstract base class for evaluating classes
 *
 * This class provides a common interface to classes that evaluate on the
 * visible sunpath. It is a collection of compute methods, that take several
 * variants to define a time span. These convert the time span to one in
 * Julian dates and pass to a low-level compute method that must be implemented
 * by the deriving class.
 *
 * This file is templated for different return types and thus header-only.
 *
 * Attention: A Julian day starts at noon!
 */
template< typename T > class Evaluator
{
public:

    //--------------------------------------------------------------------------

    /*! \brief Constructor
     *
     * The constructor store the const reference to the options.
     */
    Evaluator(const Observer& obs) : m_obs(obs)
    {

    }

    //--------------------------------------------------------------------------

    /*! \brief Evaluate for an entire year
     *
     * This method obtains the julian date of new year and sylvester of the
     * given year and passes on to the specialized method.
     *
     * \param year Year to evaluate for.
     *
     * \return Resulting value in return type of evaluator
     */
    virtual T compute(int year)
    {
        double jd0 = m_obs.getLocalJulianDate(year, 1, 1, 0, 0, 0);
        double jd1 = m_obs.getLocalJulianDate(year+1, 1, 1, 0, 0, 0);

        /* then pass on to specialized caller */
        return compute(jd0, jd1);
    }

    //--------------------------------------------------------------------------

    /*! \brief Evaluate on a given date
     *
     * This method obtains the julian date of starting and ending midnight of
     * the given day and passes on to the specialized method.
     *
     * \param day Day of date to evaluate at
     * \param month Month of date to evaluate at
     * \param year Year of date to evaluate at
     *
     * \return Resulting value in return type of evaluator
     */
    virtual T compute(unsigned int day, unsigned int month, int year)
    {
        double jd0 = m_obs.getLocalJulianDate(year, month, day, 0, 0, 0);
        double jd1 = m_obs.getLocalJulianDate(year, month, day, 24, 0, 0);

        /* then pass on to specialized caller */
        return compute(jd0, jd1);
    }

    //--------------------------------------------------------------------------

    /*! \brief Evaluate for Julian date time span
     *
     * This is the lowest-level evaluation method that takes a time span in
     * Julian dates. This must be implemented by the child class.
     *
     * \param jd0 Start Julian date, is of type double as libnova returns the
     *      Julian date as double. See Options::getLocalJulianDate. Must be
     *      local midnight.
     * \param jd1 End Julian date, is of type double as libnova returns the
     *      Julian date as double. See Options::getLocalJulianDate. Must be
     *      local midnight.
     *
     * \return Sunshine duration in days or a negative number if an error occurred.
     */
    virtual T compute(double, double) = 0;

    //--------------------------------------------------------------------------

    /*! \brief Evaluate for an entire year (in flat country, ignoring any mountains)
     *
     * This method obtains the julian date of new year and sylvester of the
     * given year and passes on to the specialized method.
     *
     * \param year Year to evaluate for.
     *
     * \return Resulting value in return type of evaluator
     */
    virtual T computePlain(int year)
    {
        double jd0 = m_obs.getLocalJulianDate(year, 1, 1, 0, 0, 0);
        double jd1 = m_obs.getLocalJulianDate(year+1, 1, 1, 0, 0, 0);

        /* then pass on to specialized caller */
        return computePlain(jd0, jd1);
    }

    //--------------------------------------------------------------------------

    /*! \brief Evaluate on a given date (in flat country, ignoring any mountains)
     *
     * This method obtains the julian date of starting and ending midnight of
     * the given day and passes on to the specialized method.
     *
     * \param day Day of date to evaluate at
     * \param month Month of date to evaluate at
     * \param year Year of date to evaluate at
     *
     * \return Resulting value in return type of evaluator
     */
    virtual T computePlain(unsigned int day, unsigned int month, int year)
    {
        double jd0 = m_obs.getLocalJulianDate(year, month, day, 0, 0, 0);
        double jd1 = m_obs.getLocalJulianDate(year, month, day, 24, 0, 0);

        /* then pass on to specialized caller */
        return computePlain(jd0, jd1);
    }

    //--------------------------------------------------------------------------

    /*! \brief Evaluate for Julian date time span (in flat country, ignoring any mountains)
     *
     * This is the lowest-level evaluation method that takes a time span in
     * Julian dates. This must be implemented by the child class.
     *
     * \param jd0 Start Julian date, is of type double as libnova returns the
     *      Julian date as double. See Options::getLocalJulianDate. Must be
     *      local midnight.
     * \param jd1 End Julian date, is of type double as libnova returns the
     *      Julian date as double. See Options::getLocalJulianDate. Must be
     *      local midnight.
     *
     * \return Sunshine duration in days or a negative number if an error occurred.
     */
    virtual T computePlain(double, double) = 0;

    //--------------------------------------------------------------------------

protected:

    /*! private const reference to the main options object */
    const Observer& m_obs;

};

#endif // EVALUATOR_H
