/*
 * (c) 2014 Rechenraum e.U. (Simon Flöry; simon.floery@rechenraum.com)
 *
 * This file is part of project oe3d (www.oe3d.at).
 *
 * oe3d is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * oe3d is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with oe3d. If not, see <http://www.gnu.org/licenses/>.
 */

#include <iostream>
#include <cmath>

#include <libnova/solar.h>

#include "EvaluatorInsolation.h"

//------------------------------------------------------------------------------

/******************************************************************************
 *
 * class EvaluatorInsolation
 *
 ******************************************************************************/

//------------------------------------------------------------------------------

/*! \brief Constructor
 *
 * The constructor store the const reference to the options.
 *
 * \param options Const reference to the options.
 */
EvaluatorInsolation::EvaluatorInsolation(const Observer& obs) :
    EvaluatorLabeledSunpath(obs),
    m_panelAzimuth(0.),
    m_panelInclination(40)
{

}

//------------------------------------------------------------------------------

/*! \brief Load labeled sunpath
 *
 * Load the labeled sunpath.
 *
 * \param sunpath Const pointer to a labeled sunpath object. We will do a
 *      shallow copy.
 *
 * \return True on success, false otherwise.
 */
bool EvaluatorInsolation::load(const LabeledSunpath* sunpath)
{
    if (sunpath->getSunpath().empty()) {
        std::cerr << "EvaluatorInsolation::load - Labeled sunpath is empty." << std::endl;
        return false;
    }

    m_sunpath = sunpath;

    return true;
}

//------------------------------------------------------------------------------

/*! \brief Evaluate for Julian date time span
 *
 *
 * \param jd0 Start Julian date, is of type double as libnova returns the
 *  Julian date as double. See Options::getLocalJulianDate. Must be
 *  local midnight.
 * \param jd1 End Julian date, is of type double as libnova returns the
 *  Julian date as double. See Options::getLocalJulianDate. Must be
 *  local midnight.
 *
 * \return Sunshine duration in days or a negative number if an error occurred.
 */
GeCore::GCReal EvaluatorInsolation::compute(double jd0, double jd1)
{
    return _compute(jd0, jd1, true);
}

//------------------------------------------------------------------------------

/*! \brief Evaluate for Julian date time span (in flat country, ignoring any mountains)
 *
 *
 * \param jd0 Start Julian date, is of type double as libnova returns the
 *  Julian date as double. See Options::getLocalJulianDate. Must be
 *  local midnight.
 * \param jd1 End Julian date, is of type double as libnova returns the
 *  Julian date as double. See Options::getLocalJulianDate. Must be
 *  local midnight.
 *
 * \return Sunshine duration in days or a negative number if an error occurred.
 */
GeCore::GCReal EvaluatorInsolation::computePlain(double jd0, double jd1)
{
    return _compute(jd0, jd1, false);
}

//------------------------------------------------------------------------------

/*! \brief Evaluate for Julian date time span
 *
 *
 * \param jd0 Start Julian date, is of type double as libnova returns the
 *  Julian date as double. See Options::getLocalJulianDate. Must be
 *  local midnight.
 * \param jd1 End Julian date, is of type double as libnova returns the
 *  Julian date as double. See Options::getLocalJulianDate. Must be
 *  local midnight.
 * \param If true, only segments of the sunpath visible (against the real
 *  horizon) are considered. If false, only segments visible against an ideal
 *  flat horizon are used.
 *
 * \return Sunshine duration in days or a negative number if an error occurred.
 */
GeCore::GCReal EvaluatorInsolation::_compute(double jd0, double jd1, bool onlyVisible)
{
    /* compute the unit length normal of the absorbing panel. */
    GeCore::GC3D nPanel(
        sin(m_panelAzimuth / 180. * M_PI) * sin(m_panelInclination / 180. * M_PI),
        cos(m_panelAzimuth / 180. * M_PI) * sin(m_panelInclination / 180. * M_PI),
        cos(m_panelInclination / 180. * M_PI));

//    std::cout << nPanel.transpose() << std::endl;

    /* Iterate through labeled sunpath */
    const std::list< LabeledSunpath::spSampleT >& s = m_sunpath->getSunpath();
    std::list< LabeledSunpath::spSampleT >::const_iterator sIt1 = s.begin();
    GeCore::GCReal insolation = 0;
    /* all segments that don't have an end point below the horizon are used */
    for ( sIt1++ ; sIt1 != s.end() ; sIt1++ ) {
        std::list< LabeledSunpath::spSampleT >::const_iterator sIt0 = sIt1;
        std::advance(sIt0, -1);

        /* This needs improvement if we ever compute in regions where there is
         * no sunset.
         */
        if (sIt0->jd < jd0-GCREAL_EPS) continue;
        if (sIt1->jd > jd1+GCREAL_EPS) break;

        if (onlyVisible) {
            /* Check if the sunpath segment is visible against real horizon */
            if (sIt0->label == LabeledSunpath::BELOW_HORIZON || sIt1->label == LabeledSunpath::BELOW_HORIZON) continue;
        } else {
            /* Check if the sunpath segment is visible against flat horizon */
            if (sIt0->p.y < LN_SOLAR_STANDART_HORIZON || sIt1->p.y < LN_SOLAR_STANDART_HORIZON) continue;
        }

        /* add up */
        insolation += integrate(*sIt0, *sIt1, nPanel, 60.);
    }

    /* We choose j0 empirically to reach the given insolation per year for
     * an average latitude for an Austrian observer. A value of 500 gives an
     * insolation of ~1300 kWh/m^2 which seems reasonable compared maps for
     * Southern Germany.
     */
    const GeCore::GCReal j0 = 500;

    /* multiply by unbiased insolation and convert to kWh/m^2 */
    return insolation * j0 / 1000.;
}

//------------------------------------------------------------------------------

/*! \brief Integrate inclination angle
 *
 * This method integrates the changing inclination angle between the two
 * provided samples of the labeled sunpath. Sun's azimuth and altitude angles
 * are linearly interpolated. The integral is approximated by the trapezoidal
 * rule.
 *
 * We write nPanel for the absorbing panel's unit normal and n(t) for the
 * flipped, normalized incoming light's direction vector (so that it is
 * actually pointing outwards). We obtainS
 *       sin(beta) = cos(pi/2 - beta) = n(t) * nPanel
 *
 * \param p0 Starting sample on a labeled sunpath
 * \param p1 Ending sample on a labeled sunpath
 * \param nPanel Normal of the absorbing panel
 * \param h Step width for the trapezoidal rule (unit is seconds)
 *
 * \return Integral of the changing inclination angle (unit is hours)
 */
GeCore::GCReal EvaluatorInsolation::integrate(
        const LabeledSunpath::spSampleT& p0,
        const LabeledSunpath::spSampleT& p1,
        const GeCore::GC3D& nPanel,
        GeCore::GCReal h) const
{
    /* Check if the sunpath is valid */
#ifdef GC_DEBUG
    if (p0.label == LabeledSunpath::UNKNOWN || p1.label == LabeledSunpath::UNKNOWN) {
        std::cerr << "EvaluatorInsolation::integrate - Invalid labaled sunpath has UNKNOWN labels." << std::endl;
        return 0.;
    }

    if (p0.jd >= p1.jd) {
        std::cerr << "EvaluatorInsolation::integrate - Invalid time ordering on sunpath." << std::endl;
        return 0.;
    }
#endif // GC_DEBUG

    /* Transform so that p0.jd is t=0, p1.jd is t=1 and h is the step size in
     * this scaled space. h is in seconds, so we need to transform it to
     * Julian days and then divide by the difference of p1.jd and p0.jd.
     */
    h /= (24*60*60);
    h /= (p1.jd - p0.jd);

    /*
     * Integral approximation by trapezoidal rule
     */

    GeCore::GCReal intVal = 0.;

    /* some constant we will need (in RADIANS) */
    GeCore::GCReal az0 = p0.p.x / 180. * M_PI;
    GeCore::GCReal alt0 = p0.p.y / 180. * M_PI;
    GeCore::GCReal az1 = p1.p.x / 180. * M_PI;
    GeCore::GCReal alt1 = p1.p.y / 180. * M_PI;

    /* Function value in starting point */
    GeCore::GC3D n0(
        sin(az0) * cos(alt0),
        cos(az0) * cos(alt0),
        sin(alt0));
    GeCore::GCReal fNew = n0.dot(nPanel);

    /* iterate over entire interval */
    for ( GeCore::GCReal t = h ; t < 1.+h/2 ; t+=h ) {
        /* the old right value is the new left value */
        GeCore::GCReal fOld = fNew;

        /* obtain azimuth and altitude through linear interpolation */
        GeCore::GCReal az = (1 - t) * az0 + t * az1;
        GeCore::GCReal alt = (1 - t) * alt0 + t * alt1;

        /* evaluate at new value */
        GeCore::GC3D n(
            sin(az) * cos(alt),
            cos(az) * cos(alt),
            sin(alt));
        fNew = n.dot(nPanel);

        /* add up to integral */
        intVal += std::max(0.5 * (fOld + fNew) * h, 0.);
    }

    /* scale back from unit interval to Julian days */
    intVal *= (p1.jd - p0.jd);

    /* multiply by 24 to get hours */
    return intVal * 24.;
}

//------------------------------------------------------------------------------
