/*
 * (c) 2014 Rechenraum e.U. (Simon Flöry; simon.floery@rechenraum.com)
 *
 * This file is part of project oe3d (www.oe3d.at).
 *
 * oe3d is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * oe3d is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with oe3d. If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef EVALUATOR_INSOLATION_H
#define EVALUATOR_INSOLATION_H

/* forward declaration */
class EvaluatorInsolation;

#include "EvaluatorLabeledSunpath.h"

//------------------------------------------------------------------------------

/******************************************************************************
 *
 * class EvaluatorInsolation
 *
 ******************************************************************************/

//------------------------------------------------------------------------------

/*! \brief Insolation is the amount of solar energy that reaches a point
 *
 * For more information on insolation, see
 *      http://en.wikipedia.org/wiki/Insolation
 *
 * Attention: A Julian day starts at noon!
 */
class EvaluatorInsolation : public EvaluatorLabeledSunpath< GeCore::GCReal >
{
public:
    /* calls to members declared only in a base template class are not resolved
     * when they are encountered, as their might be specializations that are
     * found later. So we end up with compute and computePlain unknown, unless
     * we explicitely bring them into scope with the following using
     * declarations. See
     *      http://womble.decadent.org.uk/c++/template-faq.html#base-lookup
     * for more information.
     */
    using Evaluator< GeCore::GCReal >::compute;
    using Evaluator< GeCore::GCReal >::computePlain;

    /* documented in source code */
    EvaluatorInsolation(const Observer&);

    /* documented in source code */
    bool load(const LabeledSunpath*);

    /* documented in source code */
    GeCore::GCReal compute(double, double);

    /* documented in source code */
    GeCore::GCReal computePlain(double, double);

protected:
    /* documented in source code */
    GeCore::GCReal _compute(double, double, bool);

    /* documented in source code */
    GeCore::GCReal integrate(
            const LabeledSunpath::spSampleT&,
            const LabeledSunpath::spSampleT&,
            const GeCore::GC3D&,
            GeCore::GCReal) const;

    /*! Azimuth angle of the absorbing panel, that is the deviation from a
     * southern direction in DEGREES (0=S, 90=W, 180=N, 270=E). This value to
     * defaults to zero, aligning with south.
     */
    GeCore::GCReal m_panelAzimuth;

    /*! Inclination angle of the absorbing panel in DEGREES. An angle of zero
     * means the panel is lying in the fundamental plane. An angle of 90 degrees
     * means the panel is in upright position, e.g. on a facade. This value
     * defaults to 40 degrees.
     */
    GeCore::GCReal m_panelInclination;

};

#endif // EVALUATOR_INSOLATION_H
