/*
 * (c) 2014 Rechenraum e.U. (Simon Flöry; simon.floery@rechenraum.com)
 *
 * This file is part of project oe3d (www.oe3d.at).
 *
 * oe3d is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * oe3d is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with oe3d. If not, see <http://www.gnu.org/licenses/>.
 */

#include <iostream>

#include "EvaluatorSunshineDuration.h"

//------------------------------------------------------------------------------

/******************************************************************************
 *
 * class EvaluatorSunshineDuration
 *
 ******************************************************************************/

//------------------------------------------------------------------------------

/*! \brief Constructor
 *
 * The constructor store the const reference to the options.
 *
 * \param options Const reference to the options.
 */
EvaluatorSunshineDuration::EvaluatorSunshineDuration(const Observer& obs) :
    EvaluatorLabeledSunpath(obs)
{

}

//------------------------------------------------------------------------------

/*! \brief Load labeled sunpath
 *
 * Load the labeled sunpath.
 *
 * \param sunpath Const pointer to a labeled sunpath object. We will do a
 *      shallow copy.
 *
 * \return True on success, false otherwise.
 */
bool EvaluatorSunshineDuration::load(const LabeledSunpath* sunpath)
{
    if (sunpath->getSunpath().empty()) {
        std::cerr << "EvaluatorSunshineDuration::load - Labeled sunpath is empty." << std::endl;
        return false;
    }

    m_sunpath = sunpath;

    return true;
}

//------------------------------------------------------------------------------

/*! \brief Evaluate for Julian date time span
 *
 * This method takes the labeled sun path intersected with the horizon and
 * sums up the sunny times.
 *
 * We need to map the requested time span somehow to the time span available
 * through the labeled sunpath. This is done as follows: All samples of before
 * jd0 in time are ignored, the same holds for samples larger than jd1. If there
 * is a real night, there will be no problems. If there is no night, there will
 * be some error.
 *
 * \param jd0 Start Julian date, is of type double as libnova returns the
 *  Julian date as double. See Options::getLocalJulianDate. Must be
 *  local midnight.
 * \param jd1 End Julian date, is of type double as libnova returns the
 *  Julian date as double. See Options::getLocalJulianDate. Must be
 *  local midnight.
 *
 * \return Sunshine duration in days or a negative number if an error occurred.
 */
GeCore::GCReal EvaluatorSunshineDuration::compute(double jd0, double jd1)
{
    const std::list< LabeledSunpath::spSampleT >& s = m_sunpath->getSunpath();
    std::list< LabeledSunpath::spSampleT >::const_iterator sIt1 = s.begin();
    GeCore::GCReal sunnyTime = 0;
    /* all segments that don't have an end point below the horizon are used */
    for ( sIt1++ ; sIt1 != s.end() ; sIt1++ ) {
        std::list< LabeledSunpath::spSampleT >::const_iterator sIt0 = sIt1;
        std::advance(sIt0, -1);

        /* This needs improvement if we ever compute in regions where there is
         * no sunset.
         */
        if (sIt0->jd < jd0-GCREAL_EPS) continue;
        if (sIt1->jd > jd1+GCREAL_EPS) break;

        if (sIt0->label != LabeledSunpath::BELOW_HORIZON && sIt1->label != LabeledSunpath::BELOW_HORIZON)
            sunnyTime += sIt1->jd - sIt0->jd;
    }

    return sunnyTime;
}

//------------------------------------------------------------------------------

/*! \brief Evaluate for Julian date time span (in flat country, ignoring any mountains)
 *
 * This method computes the sunshine duration for a flat horizon, e.g. at a
 * slightly negative altitude angle to account for refraction. In libnova, this
 * alitude angle is called LN_SOLAR_STANDART_HORIZON and is -0.8333 degrees.
 * This method calls libnova internally.
 *
 * \param jd0 Start Julian date, is of type double as libnova returns the
 *  Julian date as double. See Options::getLocalJulianDate. Must be
 *  local midnight.
 * \param jd1 End Julian date, is of type double as libnova returns the
 *  Julian date as double. See Options::getLocalJulianDate. Must be
 *  local midnight.
 *
 * \return Sunshine duration in days or a negative number if an error occurred.
 */
GeCore::GCReal EvaluatorSunshineDuration::computePlain(double jd0, double jd1)
{
    GeCore::GCReal sunnyTimes = 0;
    /* Would it be possible to speed this up, as we query subsequent days, or
     * is it tricky enough to justify calling Options::getSolarRST()?
     */
    for ( double jd = jd0 ; jd < jd1-0.1 ; jd += 1. ) {
        ln_rst_time rst;
        int ret = m_obs.getSolarRST(jd+0.1, rst);
        if (ret == 0)
        {
            sunnyTimes += rst.set - rst.rise;
        }
        else if (ret == 1)
        {
            sunnyTimes += 1.;
        }
    }

    return sunnyTimes;
}

//------------------------------------------------------------------------------

/*! \brief Convert Julian date to number of hours since JD=0
 *
 * We compute the sunshine duration as sum of Julian day differences, thus
 * actually the sunshine duration in days (plus fractional part). Here, we
 * convert to hours.
 *
 * \param jd Days of sunshine
 *
 * \return Hours of sunshine
 */
GeCore::GCReal EvaluatorSunshineDuration::toHours(double jd)
{
    /* yes, it is that easy. */
    return 24*jd;
}

//------------------------------------------------------------------------------
